@@ -116,6 +116,7 @@ char * gdm_locale_to_utf8 (const char *text);
 char * gdm_locale_from_utf8 (const char *text);
 
 void gdm_saveenv (void);
+const char * gdm_saved_getenv (const char *var);
 /* leaks */
 void gdm_restoreenv (void);
 
@@ -153,6 +154,8 @@ void gdm_sleep_no_signal (int secs);
  * <dir> "/" <name> <extension>
  */
 char * gdm_make_filename (const char *dir, const char *name, const char *extension);
+char * gdm_ensure_extension (const char *name, const char *extension);
+char * gdm_strip_extension (const char *name, const char *extension);
 
 #endif /* GDM_MISC_H */
 
