@@ -20,7 +20,11 @@
 /* FIXME: hack */
 extern GreeterItemInfo *welcome_string_info;
 
+/* evil globals */
 static char *file_search_path = NULL;
+static GList *button_stack = NULL;
+
+static GHashTable *pixbuf_hash = NULL;
 
 GHashTable *item_hash = NULL;
 GList *custom_items = NULL;
@@ -30,11 +34,33 @@ static gboolean parse_items (xmlNodePtr       node,
 			     GreeterItemInfo *parent,
 			     GError         **error);
 
+static GdkPixbuf *
+load_pixbuf (const char *fname, GError **error)
+{
+  GdkPixbuf *pb;
+
+  if (pixbuf_hash == NULL)
+    pixbuf_hash = g_hash_table_new_full (g_str_hash,
+					 g_str_equal,
+					 g_free,
+					 (GDestroyNotify)g_object_unref);
+  pb = g_hash_table_lookup (pixbuf_hash, fname);
+  if (pb != NULL)
+    return g_object_ref (pb);
+
+  pb = gdk_pixbuf_new_from_file (fname, error);
+  if G_UNLIKELY (pb == NULL)
+    return NULL;
+
+  g_hash_table_insert (pixbuf_hash, g_strdup (fname), g_object_ref (pb));
+
+  return pb;
+}
 
 GQuark
 greeter_parser_error_quark (void)
 {
-  static GQuark quark;
+  static GQuark quark = 0;
   if (!quark)
     quark = g_quark_from_static_string ("greeter_parser_error");
 
@@ -93,7 +119,7 @@ parse_button (xmlNodePtr node,
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "bad button spec %s\n", prop);
+		       "bad button spec %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -142,7 +168,7 @@ parse_pos (xmlNodePtr       node,
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Unknown anchor type %s\n", prop);
+		       "Unknown anchor type %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -154,16 +180,21 @@ parse_pos (xmlNodePtr       node,
   if (prop)
     {
       info->x = g_ascii_strtod (prop, &p);
-      
+
       if ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad position specifier %s\n", prop);
+		       "Bad position specifier %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
+
+      if (prop[0] == '-' || info->x < 0)
+        info->x_negative = TRUE;
+      else
+        info->x_negative = FALSE;
       
       if (strchr (prop, '%') != NULL)
 	info->x_type = GREETER_ITEM_POS_RELATIVE;
@@ -182,10 +213,15 @@ parse_pos (xmlNodePtr       node,
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad position specifier %s\n", prop);
+		       "Bad position specifier %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
+
+      if (prop[0] == '-' || info->y < 0)
+        info->y_negative = TRUE;
+      else
+        info->y_negative = FALSE;
       
       if (strchr (prop, '%') != NULL)
 	info->y_type = GREETER_ITEM_POS_RELATIVE;
@@ -202,13 +238,13 @@ parse_pos (xmlNodePtr       node,
       else
 	{
 	  info->width = g_ascii_strtod (prop, &p);
-      
+
 	  if ((char *)prop == p)
 	    {
 	      g_set_error (error,
 			   GREETER_PARSER_ERROR,
 			   GREETER_PARSER_ERROR_BAD_SPEC,
-			   "Bad size specifier %s\n", prop);
+			   "Bad size specifier %s", prop);
 	      xmlFree (prop);
 	      return FALSE;
 	    }
@@ -235,7 +271,7 @@ parse_pos (xmlNodePtr       node,
 	      g_set_error (error,
 			   GREETER_PARSER_ERROR,
 			   GREETER_PARSER_ERROR_BAD_SPEC,
-			   "Bad size specifier %s\n", prop);
+			   "Bad size specifier %s", prop);
 	      xmlFree (prop);
 	      return FALSE;
 	    }
@@ -264,7 +300,7 @@ parse_pos (xmlNodePtr       node,
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad expand spec %s\n", prop);
+		       "Bad expand spec %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -294,7 +330,7 @@ parse_show (xmlNodePtr       node,
 
   /* Note: subtype is deprecated, use type only */
   prop = xmlGetProp (node, "subtype");
-  if (prop != NULL)
+  if G_UNLIKELY (prop != NULL)
     {
       /* code for legacy uses of subtype only, are there any such
        * themes out there?  The Bluecurve was the one this was made
@@ -403,7 +439,7 @@ parse_box (xmlNodePtr       node,
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad orientation %s\n", prop);
+		       "Bad orientation %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -427,7 +463,7 @@ parse_box (xmlNodePtr       node,
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad homogenous spec %s\n", prop);
+		       "Bad homogenous spec %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -441,12 +477,12 @@ parse_box (xmlNodePtr       node,
     {
       info->box_x_padding = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad padding specification %s\n", prop);
+		       "Bad padding specification %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -458,12 +494,12 @@ parse_box (xmlNodePtr       node,
     {
       info->box_y_padding = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad padding specification %s\n", prop);
+		       "Bad padding specification %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -475,12 +511,12 @@ parse_box (xmlNodePtr       node,
     {
       info->box_min_width = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad min-width specification %s\n", prop);
+		       "Bad min-width specification %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -492,12 +528,12 @@ parse_box (xmlNodePtr       node,
     {
       info->box_min_height = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad min-height specification %s\n", prop);
+		       "Bad min-height specification %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -509,12 +545,12 @@ parse_box (xmlNodePtr       node,
     {
       info->box_spacing = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad spacing specification %s\n", prop);
+		       "Bad spacing specification %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -536,20 +572,20 @@ parse_color (const char *str,
 {
   guint32 col;
   int i;
-  if (str[0] != '#')
+  if G_UNLIKELY (str[0] != '#')
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_BAD_SPEC,
-		   "colors must start with #, %s is an invalid color\n", str);
+		   "colors must start with #, %s is an invalid color", str);
       return FALSE;
     }
-  if (strlen (str) != 7)
+  if G_UNLIKELY (strlen (str) != 7)
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_BAD_SPEC,
-		   "Colors must be on the format #xxxxxx, %s is an invalid color\n", str);
+		   "Colors must be on the format #xxxxxx, %s is an invalid color", str);
       return FALSE;
     }
 
@@ -564,21 +600,23 @@ parse_color (const char *str,
 }
 
 static gboolean
-parse_state_file (xmlNodePtr node,
-		  GreeterItemInfo  *info,
-		  GreeterItemState state,
-		  GError         **error)
+parse_state_file_pixmap (xmlNodePtr node,
+			 GreeterItemInfo  *info,
+			 GreeterItemState state,
+			 GError         **error)
 {
   xmlChar *prop;
   char *p;
+
+  info->have_state |= (1<<state);
   
   prop = xmlGetProp (node, "file");
   if (prop)
     {
       if (g_path_is_absolute (prop))
-	info->files[state] = g_strdup (prop);
+	info->data.pixmap.files[state] = g_strdup (prop);
       else
-	info->files[state] = g_build_filename (file_search_path,
+	info->data.pixmap.files[state] = g_build_filename (file_search_path,
 					       prop,
 					       NULL);
 	
@@ -588,65 +626,81 @@ parse_state_file (xmlNodePtr node,
   prop = xmlGetProp (node, "tint");
   if (prop)
     {
-      if (!parse_color (prop, &info->tints[state], error))
+      if (!parse_color (prop, &info->data.pixmap.tints[state], error))
 	return FALSE;
-      info->have_tint[state] = TRUE;
+      info->data.pixmap.have_tint |= (1<<state);
       xmlFree (prop);
     }
 
   prop = xmlGetProp (node, "alpha");
   if (prop)
     {
-      info->alphas[state] = g_ascii_strtod (prop, &p);
+      double alpha = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad alpha specifier format %s\n", prop);
+		       "Bad alpha specifier format %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
       xmlFree (prop);
+
+      if (alpha >= 1.0)
+        info->data.pixmap.alphas[state] = 0xff;
+      else if (alpha < 0)
+        info->data.pixmap.alphas[state] = 0;
+      else
+        info->data.pixmap.alphas[state] = floor (alpha * 0xff);
     }
   
   return TRUE;
 }
 
 static gboolean
-parse_state_color (xmlNodePtr node,
-		   GreeterItemInfo  *info,
-		   GreeterItemState state,
-		   GError         **error)
+parse_state_color_rect (xmlNodePtr node,
+			GreeterItemInfo  *info,
+			GreeterItemState state,
+			GError         **error)
 {
   xmlChar *prop;
   char *p;
+
+  info->have_state |= (1<<state);
   
   prop = xmlGetProp (node, "color");
   if (prop)
     {
-      if (!parse_color (prop, &info->colors[state], error))
+      if G_UNLIKELY (!parse_color (prop, &info->data.rect.colors[state], error))
 	return FALSE;
-      info->have_color[state] = TRUE;
+      info->data.rect.have_color |= (1<<state);
       xmlFree (prop);
     }
 
   prop = xmlGetProp (node, "alpha");
   if (prop)
     {
-      info->alphas[state] = g_ascii_strtod (prop, &p);
+      double alpha = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad alpha specifier format %s\n", prop);
+		       "Bad alpha specifier format %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
       xmlFree (prop);
+
+      if (alpha >= 1.0)
+        info->data.rect.alphas[state] = 0xff;
+      else if (alpha < 0)
+        info->data.rect.alphas[state] = 0;
+      else
+        info->data.rect.alphas[state] = floor (alpha * 0xff);
     }
   
   return TRUE;
@@ -667,49 +721,49 @@ parse_pixmap (xmlNodePtr        node,
     {
       if (strcmp (child->name, "normal") == 0)
 	{
-	  if (!parse_state_file (child, info, GREETER_ITEM_STATE_NORMAL, error))
+	  if G_UNLIKELY (!parse_state_file_pixmap (child, info, GREETER_ITEM_STATE_NORMAL, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "prelight") == 0)
 	{
-	  if (!parse_state_file (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
+	  if G_UNLIKELY (!parse_state_file_pixmap (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "active") == 0)
 	{
-	  if (!parse_state_file (child, info, GREETER_ITEM_STATE_ACTIVE, error))
+	  if G_UNLIKELY (!parse_state_file_pixmap (child, info, GREETER_ITEM_STATE_ACTIVE, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0)
 	{
-	  if (!parse_fixed (child, info, error))
+	  if G_UNLIKELY (!parse_fixed (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "box") == 0)
 	{
-	  if (!parse_box (child, info, error))
+	  if G_UNLIKELY (!parse_box (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       
       child = child->next;
     }
 
-  if (!info->files[GREETER_ITEM_STATE_NORMAL])
+  if G_UNLIKELY (!info->data.pixmap.files[GREETER_ITEM_STATE_NORMAL])
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_BAD_SPEC,
-		   "No filename specified for normal state\n");
+		   "No filename specified for normal state");
       return FALSE;
     }
   
@@ -717,15 +771,15 @@ parse_pixmap (xmlNodePtr        node,
     {
       for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
 	{
-	  if (info->files[i] != NULL)
+	  if (info->data.pixmap.files[i] != NULL)
 	    {
-	      info->orig_pixbufs[i] = gdk_pixbuf_new_from_file (info->files[i], error);
+	      info->data.pixmap.pixbufs[i] = load_pixbuf (info->data.pixmap.files[i], error);
 	      
-	      if (info->orig_pixbufs[i] == NULL)
+	      if G_UNLIKELY (info->data.pixmap.pixbufs[i] == NULL)
 		return FALSE;
 	    }
 	  else
-	    info->orig_pixbufs[i] = NULL;
+	    info->data.pixmap.pixbufs[i] = NULL;
 	}
     }
 
@@ -746,37 +800,37 @@ parse_rect (xmlNodePtr node,
     {
       if (strcmp (child->name, "normal") == 0)
 	{
-	  if (!parse_state_color (child, info, GREETER_ITEM_STATE_NORMAL, error))
+	  if G_UNLIKELY (!parse_state_color_rect (child, info, GREETER_ITEM_STATE_NORMAL, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "prelight") == 0)
 	{
-	  if (!parse_state_color (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
+	  if G_UNLIKELY (!parse_state_color_rect (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "active") == 0)
 	{
-	  if (!parse_state_color (child, info, GREETER_ITEM_STATE_ACTIVE, error))
+	  if G_UNLIKELY (!parse_state_color_rect (child, info, GREETER_ITEM_STATE_ACTIVE, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0)
 	{
-	  if (!parse_fixed (child, info, error))
+	  if G_UNLIKELY (!parse_fixed (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "box") == 0)
 	{
-	  if (!parse_box (child, info, error))
+	  if G_UNLIKELY (!parse_box (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       
@@ -785,15 +839,10 @@ parse_rect (xmlNodePtr node,
 
   for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
     {
-      if (!info->have_color[i])
+      if ( ! (info->data.rect.have_color & (1<<i)))
 	continue;
       
-      if (info->alphas[i] >= 1.0)
-	info->colors[i] = (info->colors[i] << 8) | 0xff;
-      else if (info->alphas[i] > 0)
-	info->colors[i] = (info->colors[i] << 8) | (guint) floor (0xff*info->alphas[i]);
-      else
-	info->colors[i] = 0;
+      info->data.rect.colors[i] = (info->data.rect.colors[i] << 8) | (guint) info->data.rect.alphas[i];
     }
   
   return TRUE;
@@ -809,16 +858,18 @@ parse_state_text (xmlNodePtr node,
   xmlChar *prop;
   char *p;
 
+  info->have_state |= (1<<state);
+
   prop = xmlGetProp (node, "font");
   if (prop)
     {
-      info->fonts[state] = pango_font_description_from_string (prop);
-      if (info->fonts[state] == NULL)
+      info->data.text.fonts[state] = pango_font_description_from_string (prop);
+      if G_UNLIKELY (info->data.text.fonts[state] == NULL)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad font specification %s\n", prop);
+		       "Bad font specification %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -828,27 +879,34 @@ parse_state_text (xmlNodePtr node,
   prop = xmlGetProp (node, "color");
   if (prop)
     {
-      if (!parse_color (prop, &info->colors[state], error))
+      if G_UNLIKELY (!parse_color (prop, &info->data.text.colors[state], error))
 	return FALSE;
-      info->have_color[state] = TRUE;
+      info->data.text.have_color |= (1<<state);
       xmlFree (prop);
    }
 
   prop = xmlGetProp (node, "alpha");
   if (prop)
     {
-      info->alphas[state] = g_ascii_strtod (prop, &p);
+      double alpha = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad alpha specifier format %s\n", prop);
+		       "Bad alpha specifier format %s", prop);
 	  xmlFree (prop);
 	  return FALSE;
 	}
       xmlFree (prop);
+
+      if (alpha >= 1.0)
+        info->data.rect.alphas[state] = 0xff;
+      else if (alpha < 0)
+        info->data.rect.alphas[state] = 0;
+      else
+        info->data.rect.alphas[state] = floor (alpha * 0xff);
     }
   
   return TRUE;
@@ -1007,7 +1065,7 @@ parse_stock (xmlNodePtr node,
           g_set_error (error,
 		       GREETER_PARSER_ERROR,
 		       GREETER_PARSER_ERROR_BAD_SPEC,
-		       "Bad stock label type\n");
+		       "Bad stock label type");
 	  xmlFree (prop);
 	  return FALSE;
 	}
@@ -1024,11 +1082,83 @@ parse_stock (xmlNodePtr node,
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_BAD_SPEC,
-		   "Stock type not specified\n");
+		   "Stock type not specified");
       return FALSE;
     }
 }
 
+static void
+do_font_size_reduction (GreeterItemInfo *info)
+{
+  double size_reduction = 1.0;
+  int i;
+
+  if (gdm_wm_screen.width <= 800 &&
+      gdm_wm_screen.width > 640)
+    size_reduction = PANGO_SCALE_SMALL;
+  else if (gdm_wm_screen.width <= 640)
+    size_reduction = PANGO_SCALE_X_SMALL;
+
+  if (size_reduction < 0.99)
+    {
+      for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
+        {
+          if (info->data.text.fonts[i] != NULL)
+	    {
+	      int old_size = pango_font_description_get_size (info->data.text.fonts[i]);
+	      pango_font_description_set_size (info->data.text.fonts[i], old_size * size_reduction);
+	    }
+	}
+    }
+}
+
+
+static gboolean
+parse_label_pos_extras (xmlNodePtr       node,
+			GreeterItemInfo *info,
+			GError         **error)
+{
+  xmlChar *prop;
+  char *p;
+  
+  prop = xmlGetProp (node, "max-width");
+  if (prop)
+    {
+      info->data.text.max_width = g_ascii_strtod (prop, &p);
+      
+      if G_UNLIKELY ((char *)prop == p)
+	{
+	  g_set_error (error,
+		       GREETER_PARSER_ERROR,
+		       GREETER_PARSER_ERROR_BAD_SPEC,
+		       "Bad max-width specification %s", prop);
+	  xmlFree (prop);
+	  return FALSE;
+	}
+      xmlFree (prop);
+    }
+
+  prop = xmlGetProp (node, "max-screen-percent-width");
+  if (prop)
+    {
+      info->data.text.max_screen_percent_width = g_ascii_strtod (prop, &p);
+      
+      if G_UNLIKELY ((char *)prop == p)
+	{
+	  g_set_error (error,
+		       GREETER_PARSER_ERROR,
+		       GREETER_PARSER_ERROR_BAD_SPEC,
+		       "Bad max-screen-percent-width specification %s", prop);
+	  xmlFree (prop);
+	  return FALSE;
+	}
+      xmlFree (prop);
+    }
+
+  return TRUE;
+}
+  
+
 static gboolean
 parse_label (xmlNodePtr        node,
 	     GreeterItemInfo  *info,
@@ -1036,45 +1166,51 @@ parse_label (xmlNodePtr        node,
 {
   xmlNodePtr child;
   int i;
-  double size_reduction = 1.0;
   char *translated_text;
   gint translation_score = 1000;
   
   translated_text = NULL;
-  
+
   child = node->children;
   while (child)
     {
       if (strcmp (child->name, "normal") == 0)
 	{
-	  if (!parse_state_text (child, info, GREETER_ITEM_STATE_NORMAL, error))
+	  if G_UNLIKELY (!parse_state_text (child, info, GREETER_ITEM_STATE_NORMAL, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "prelight") == 0)
 	{
-	  if (!parse_state_text (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
+	  if G_UNLIKELY (!parse_state_text (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
+	    return FALSE;
+	}
+      else if (strcmp (child->name, "active") == 0)
+	{
+	  if G_UNLIKELY (!parse_state_text (child, info, GREETER_ITEM_STATE_ACTIVE, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
+	    return FALSE;
+	  if G_UNLIKELY (!parse_label_pos_extras (child, info, error))
 	    return FALSE;
 	}
       else if (child->type == XML_ELEMENT_NODE &&
 	       strcmp (child->name, "text") == 0)
 	{
-	  if (!parse_translated_text (child, &translated_text, &translation_score, error))
+	  if G_UNLIKELY (!parse_translated_text (child, &translated_text, &translation_score, error))
 	    return FALSE;
 	}
       else if (child->type == XML_ELEMENT_NODE &&
 	       strcmp (child->name, "stock") == 0)
 	{
-	  if (!parse_stock (child, info, &translated_text, &translation_score, error))
+	  if G_UNLIKELY (!parse_stock (child, info, &translated_text, &translation_score, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0 ||
@@ -1109,39 +1245,18 @@ parse_label (xmlNodePtr        node,
 
   for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
     {
-      if (!info->have_color[i])
+      if ( ! (info->data.text.have_color & (1<<i)))
 	continue;
       
-      if (info->alphas[i] >= 1.0)
-	info->colors[i] = (info->colors[i] << 8) | 0xff;
-      else if (info->alphas[i] > 0)
-	info->colors[i] = (info->colors[i] << 8) | (guint) floor (0xff*info->alphas[i]);
-      else
-	info->colors[i] = 0;
+      info->data.text.colors[i] = (info->data.text.colors[i] << 8) | (guint) info->data.text.alphas[i];
     }
   
-  if (info->fonts[GREETER_ITEM_STATE_NORMAL] == NULL)
-    info->fonts[GREETER_ITEM_STATE_NORMAL] = pango_font_description_from_string ("Sans");
+  if (info->data.text.fonts[GREETER_ITEM_STATE_NORMAL] == NULL)
+    info->data.text.fonts[GREETER_ITEM_STATE_NORMAL] = pango_font_description_from_string ("Sans");
 
-  if (gdm_wm_screen.width <= 800 &&
-      gdm_wm_screen.width > 640)
-    size_reduction = PANGO_SCALE_SMALL;
-  else if (gdm_wm_screen.width <= 640)
-    size_reduction = PANGO_SCALE_X_SMALL;
+  do_font_size_reduction (info);
 
-  if (size_reduction < 0.99)
-    {
-      for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
-        {
-          if (info->fonts[i] != NULL)
-	    {
-	      int old_size = pango_font_description_get_size (info->fonts[i]);
-	      pango_font_description_set_size (info->fonts[i], old_size * size_reduction);
-	    }
-	}
-    }
-
-  info->orig_text = translated_text;
+  info->data.text.orig_text = translated_text;
   
   return TRUE;
 }
@@ -1159,7 +1274,7 @@ parse_listitem (xmlNodePtr        node,
   
   prop = xmlGetProp (node, "id");
   
-  if (prop)
+  if G_LIKELY (prop)
     {
       li = g_new0 (GreeterItemListItem, 1);
       li->id = g_strdup (prop);
@@ -1180,7 +1295,7 @@ parse_listitem (xmlNodePtr        node,
       if (child->type == XML_ELEMENT_NODE &&
 	  strcmp (child->name, "text") == 0)
 	{
-	  if ( ! parse_translated_text (child, &translated_text, &translation_score, error))
+	  if G_UNLIKELY ( ! parse_translated_text (child, &translated_text, &translation_score, error))
 	    {
               g_free (li->id);
               g_free (li);
@@ -1191,7 +1306,7 @@ parse_listitem (xmlNodePtr        node,
       child = child->next;
     }
 
-  if (translated_text == NULL)
+  if G_UNLIKELY (translated_text == NULL)
     {
       g_free (li->id);
       g_free (li);
@@ -1203,7 +1318,7 @@ parse_listitem (xmlNodePtr        node,
     }
   li->text = translated_text;
 
-  info->list_items = g_list_append (info->list_items, li);
+  info->data.list.items = g_list_append (info->data.list.items, li);
 
   return TRUE;
 }
@@ -1220,17 +1335,17 @@ parse_list (xmlNodePtr        node,
     {
       if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "listitem") == 0)
 	{
-	  if ( ! parse_listitem (child, info, error))
+	  if G_UNLIKELY ( ! parse_listitem (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0 ||
@@ -1246,8 +1361,8 @@ parse_list (xmlNodePtr        node,
       child = child->next;
     }
 
-  if (info->list_items != NULL) {
-    if (strcmp (info->id, "userlist") == 0) {
+  if (info->data.list.items != NULL) {
+    if G_UNLIKELY (strcmp (info->id, "userlist") == 0) {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_BAD_SPEC,
@@ -1270,14 +1385,19 @@ parse_entry (xmlNodePtr        node,
   child = node->children;
   while (child)
     {
-      if (strcmp (child->name, "pos") == 0)
+      if (strcmp (child->name, "normal") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_state_text (child, info, GREETER_ITEM_STATE_NORMAL, error))
+	    return FALSE;
+	}
+      else if (strcmp (child->name, "pos") == 0)
+	{
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0 ||
@@ -1293,6 +1413,8 @@ parse_entry (xmlNodePtr        node,
       child = child->next;
     }
 
+  do_font_size_reduction (info);
+
   return TRUE;
 }
 
@@ -1318,7 +1440,7 @@ parse_items (xmlNodePtr  node,
       {
 	if (child->type == XML_ELEMENT_NODE)
 	  {
-	    if (strcmp (child->name, "item") != 0)
+	    if G_UNLIKELY (strcmp (child->name, "item") != 0)
 	      {
 		g_set_error (error,
 			     GREETER_PARSER_ERROR,
@@ -1328,7 +1450,7 @@ parse_items (xmlNodePtr  node,
 	      }
 	    
 	    type = xmlGetProp (child, "type");
-	    if (!type)
+	    if G_UNLIKELY (!type)
 	      {
 		g_set_error (error,
 			     GREETER_PARSER_ERROR,
@@ -1362,9 +1484,14 @@ parse_items (xmlNodePtr  node,
 	    info = greeter_item_info_new (parent, item_type);
 	    
 	    parse_id (child, info);
-	    if ( ! parse_button (child, info, error))
+	    if G_UNLIKELY ( ! parse_button (child, info, error))
 	      return FALSE;
 
+	    if (button_stack != NULL)
+	      info->my_button = button_stack->data;
+	    if (info->button)
+	      button_stack = g_list_prepend (button_stack, info);
+
 	    switch (item_type)
 	      {
 	      case GREETER_ITEM_TYPE_SVG:
@@ -1392,8 +1519,11 @@ parse_items (xmlNodePtr  node,
 			     "Bad item type");
 		res = FALSE;
 	      }
+
+	    if (info->button)
+	      button_stack = g_list_remove (button_stack, info);
 	    
-	    if (!res)
+	    if G_UNLIKELY (!res)
 	      return FALSE;
 
 	    items = g_list_prepend (items, info);
@@ -1434,7 +1564,7 @@ greeter_parse (const char *file, const char *datadir,
   g_free (file_search_path);
   file_search_path = g_strdup (datadir);
   
-  if (!g_file_test (file, G_FILE_TEST_EXISTS))
+  if G_UNLIKELY (!g_file_test (file, G_FILE_TEST_EXISTS))
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
@@ -1445,7 +1575,7 @@ greeter_parse (const char *file, const char *datadir,
   
 
   doc = xmlParseFile (file);
-  if (doc == NULL)
+  if G_UNLIKELY (doc == NULL)
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
@@ -1455,7 +1585,7 @@ greeter_parse (const char *file, const char *datadir,
     }
   
   node = xmlDocGetRootElement (doc);
-  if (node == NULL)
+  if G_UNLIKELY (node == NULL)
     {
       xmlFreeDoc (doc);
       g_set_error (error,
@@ -1465,7 +1595,7 @@ greeter_parse (const char *file, const char *datadir,
       return NULL;
     }
   
-  if (strcmp (node->name, "greeter") != 0)
+  if G_UNLIKELY (strcmp (node->name, "greeter") != 0)
     {
       xmlFreeDoc (doc);
       g_set_error (error,
@@ -1482,7 +1612,13 @@ greeter_parse (const char *file, const char *datadir,
 
   root = greeter_item_info_new (NULL, GREETER_ITEM_TYPE_RECT);
   res = parse_items (node, &items, root, error);
-  if (!res)
+
+  /* Now we can whack the hash, we don't want to keep cached
+     pixbufs around anymore */
+  g_hash_table_destroy (pixbuf_hash);
+  pixbuf_hash = NULL;
+
+  if G_UNLIKELY (!res)
     {
       welcome_string_info = NULL;
 
@@ -1491,6 +1627,9 @@ greeter_parse (const char *file, const char *datadir,
       g_list_free (custom_items);
       custom_items = NULL;
 
+      g_list_free (button_stack);
+      button_stack = NULL;
+
       g_list_foreach (items, (GFunc) greeter_item_info_free, NULL);
       g_list_free (items);
       items = NULL;
