@@ -25,6 +25,7 @@
 #include <time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <locale.h>
 
 char **	ve_split (const char *s);
 char *	ve_first_word (const char *s);
@@ -66,12 +67,29 @@ char *		ve_find_file			(const char *filename,
 char * ve_locale_to_utf8 (const char *str);
 char * ve_locale_from_utf8 (const char *str);
 
+/* These two functions will ALWAYS return a non-NULL string,
+ * if there is an error, they return the unconverted string */
+char * ve_filename_to_utf8 (const char *str);
+char * ve_filename_from_utf8 (const char *str);
+
 /* works with utf-8 strings and always returns something */
 char * ve_strftime (struct tm *the_tm, const char *format);
 
 /* function which doesn't stop on signals */
 pid_t ve_waitpid_no_signal (pid_t pid, int *status, int options);
 
+/* Testing for existance of a certain locale */
+gboolean ve_locale_exists (const char *loc);
 
+int ve_setenv (const char *name, const char *value, gboolean overwrite);
+void ve_unsetenv (const char *name);
+void ve_clearenv (void);
+
+/* just like gnome_i18n_get_language_list (in fact if you link to
+   gnome it uses that */
+const GList * ve_i18n_get_language_list (const gchar *category_name);
+
+char * ve_find_prog_in_path (const char *prog, const char *path);
+gboolean ve_is_prog_in_path (const char *prog, const char *path);
 
 #endif /* VE_MISC_H */
